//**********************************************************************
// CRhsEventLog
// ============
//
// Class to encapsulate the Windows NT Event Log
//
// John Rennie
// jrennie@cix.compulink.co.uk
// 28th July 1999
//**********************************************************************

#ifndef STRICT
#define STRICT
#endif

#include <windows.h>
#include <string.h>
#include "RhsEventLog.h"


//**********************************************************************
// Prototypes for some helper functions
// ------------------------------------
//**********************************************************************

static DWORD DaysPerMonth(int Month, int Year);
static DWORD SubstituteEnvironmentVariables(char* InBuf, char* OutBuf, DWORD MaxLen);


//**********************************************************************
// Global variables
// ----------------
//**********************************************************************

#define MAX_SUBSTRINGS 64


//**********************************************************************
// CRhsEventLog::CRhsEventLog
// CRhsEventLog::~CRhsEventLog
// ---------------------------
// Close log and free buffers on destruction
//**********************************************************************

CRhsEventLog::CRhsEventLog()
{
  m_hEventLog = NULL;
  m_LogType = 0;
  m_Record = NULL;
  m_Message = NULL;
}


CRhsEventLog::~CRhsEventLog()
{
  if (m_hEventLog)
    CloseEventLog(m_hEventLog);

  if (m_Record)
    free(m_Record);

  if (m_Message)
    LocalFree(m_Message); // NB m_Message is LocalAlloced
}


//**********************************************************************
// CRhsEventLog::OpenLog
// ---------------------
// Open one of the three event logs
// 1 - System
// 2 - Security
// 3 - Application
//**********************************************************************

BOOL CRhsEventLog::OpenLog(int Type, BOOL ReadBack, const char* Computer)
{

// Check the log is closed

  CloseLog();

// Open the log requested

  switch (Type)
  { 

// 1 is the system log

    case 1:
      m_LogType = 1;
      m_hEventLog = OpenEventLog(Computer, "System");
      if (!m_hEventLog)
        return FALSE;
      break;

// 2 is the security log

    case 2:
      m_LogType = 2;
      m_hEventLog = OpenEventLog(Computer, "Security");
      if (!m_hEventLog)
        return FALSE;
      break;

// 3 is the application log

    case 3:
      m_LogType = 3;
      m_hEventLog = OpenEventLog(Computer, "Application");
      if (!m_hEventLog)
        return FALSE;
      break;

// No other values are valid

    default:
      return FALSE;
  }

// Save the computer name

  strcpy(m_Computer, Computer ? Computer : "");

// Move to the first or last record as requested

  SetDirection(ReadBack);

  ReadRecord();

// All done

  return TRUE;
}


BOOL CRhsEventLog::OpenLog(char* Type, BOOL ReadBack, const char* Computer)
{
  if (strcmpi(Type, "System") == 0)
    return OpenLog(1, ReadBack, Computer);
  else if (strcmpi(Type, "Security") == 0)
    return OpenLog(2, ReadBack, Computer);
  else if (strcmpi(Type, "Application") == 0)
    return OpenLog(3, ReadBack, Computer);
  else
    return OpenLog(0, ReadBack, NULL); // Deliberately pass an invalid arg
}


//**********************************************************************
// CRhsEventLog::CloseLog
// ----------------------
// Close the event log
//**********************************************************************

void CRhsEventLog::CloseLog(void)
{
  if (m_hEventLog)
  { CloseEventLog(m_hEventLog);
    m_hEventLog = NULL;
  }
}


//**********************************************************************
// CRhsEventLog::ReadRecord
// ------------------------
// Read a single record.
// If Offset is 0 then a sequential read is done in the direction saved 
// in m_ReadDirection.  If Offset is not 0 then a SEEK is done.  The 
// default for Offset is 0 so it can simply be omitted to do sequential 
// reads,
//**********************************************************************

BOOL CRhsEventLog::ReadRecord(DWORD Offset)
{ BOOL b;
  DWORD bytes_read, bytes_needed;

// Check the log is open

  if (!m_hEventLog)
    return FALSE;

// Allocate the buffer so it's big enough for the EVENTLOGRECORD structure
// but not for the extra data.  We will reallocate it when we know what
// size is needed.

  if (m_Record)
    free(m_Record);

  m_Record = (EVENTLOGRECORD*) malloc(sizeof(EVENTLOGRECORD));

// Get the record using the current buffer size.  This will probably
// fail, but what we want is the required buffer size so we can reallocate
// it accordingly.

  b = ReadEventLog(m_hEventLog, 
                   m_ReadDirection | (Offset == 0 ? EVENTLOG_SEQUENTIAL_READ : EVENTLOG_SEEK_READ),
                   Offset,
                   m_Record,
                   sizeof(EVENTLOGRECORD),
                   &bytes_read,
                   &bytes_needed);

// If the buffer was too small then increase it

  if (!b && GetLastError() == ERROR_INSUFFICIENT_BUFFER)
  { m_Record = (EVENTLOGRECORD*) realloc(m_Record, bytes_needed);

    b = ReadEventLog(m_hEventLog, 
                     m_ReadDirection | (Offset == 0 ? EVENTLOG_SEQUENTIAL_READ : EVENTLOG_SEEK_READ),
                     Offset,
                     m_Record,
                     bytes_needed,
                     &bytes_read,
                     &bytes_needed);
  }

// If the read failed delete the buffer

  if (!b)
  { DWORD d = GetLastError();

    free(m_Record);
    m_Record = NULL;

    SetLastError(d);
    return FALSE;
  }

// All done

  return TRUE;
}


//**********************************************************************
// CRhsEventLog::Category
// --------------------
// Return the event type as a string
//**********************************************************************

const char* CRhsEventLog::Category(void)
{

// Check there is a record

  if (!m_Record)
    return NULL;

// Return the event type as a string

  switch (m_Record->EventType)
  { case EVENTLOG_ERROR_TYPE:
      return "Error";

    case EVENTLOG_WARNING_TYPE:
      return "Warning";

    case EVENTLOG_INFORMATION_TYPE:
      return "Information";

    case EVENTLOG_AUDIT_SUCCESS:
      return "Success Audit";

    case EVENTLOG_AUDIT_FAILURE:
      return "Failure Audit";

    default:
      return "None";
  }
}


//**********************************************************************
// CRhsEventLog::Source
// --------------------
// Return the name of the module that generated the event
//**********************************************************************

const char* CRhsEventLog::Source(void)
{
  return m_Record ? (char*) m_Record + sizeof(EVENTLOGRECORD) : NULL;
}


//**********************************************************************
// CRhsEventLog::Computer
// ----------------------
// Return the name of the computer that generated the event
//**********************************************************************

const char* CRhsEventLog::Computer(void)
{
  return m_Record ? (char*) m_Record + sizeof(EVENTLOGRECORD) + strlen(Source()) + 1 : NULL;
}


//**********************************************************************
// CRhsEventLog::Date
// ------------------
// Return the date/time the record was generated as a SYSTEMTIME
//**********************************************************************

BOOL CRhsEventLog::Time(SYSTEMTIME* st)
{ __int64 mstime;
  __int64 SecsTo1970 = 116444736000000000;
  FILETIME ft, lft;

// Check there is a record

  if (!m_Record)
    return FALSE;

// Convert the time to MS file time format i.e. number of 100-nanosecond
// intervals since 12:00 A.M. January 1, 1601 (UTC).

  mstime = Int32x32To64(m_Record->TimeGenerated,10000000) + SecsTo1970;

// Now convert this to a FILETIME and use WIN32 calls to convert the
// FILETIME to a SYSTEMTIME.

  ft.dwLowDateTime = (DWORD) mstime;
  ft.dwHighDateTime = (DWORD)(mstime >> 32);

  FileTimeToLocalFileTime(&ft, &lft);
  FileTimeToSystemTime(&lft, st);

// All done

  return TRUE;
}


//**********************************************************************
// CRhsEventLog::Text
// ------------------
// Retrieve the text description for the current record
//**********************************************************************

const char* CRhsEventLog::Message(void)
{ LONG l;
  HKEY hkey;
  DWORD regtype, d;
  HMODULE h_messagefile;
  char* msgfileptr;
  char* stringptr;
  char* substring[MAX_SUBSTRINGS];
  char messagefile[MAX_PATH+1], s[MAX_PATH+1];

// Check the record exists

  if (!m_Record)
    return NULL;

// Clear the current message

  if (m_Message)
  { LocalFree(m_Message);
    m_Message = NULL;
  }

// Get the message file for this event

  strcpy(s, "SYSTEM\\CurrentControlSet\\Services\\EventLog\\");

  if (m_LogType == 1)
    strcat(s, "System\\");
  else if (m_LogType == 2)
    strcat(s, "Security\\");
  else if (m_LogType == 3)
    strcat(s, "Application\\");
  else
    return NULL;

  strcat(s, (char*) m_Record + sizeof(EVENTLOGRECORD));

  l = RegOpenKeyEx(HKEY_LOCAL_MACHINE, s, 0, KEY_READ, &hkey);
  if (l != ERROR_SUCCESS)
    return NULL;

  d = MAX_PATH;
  l = RegQueryValueEx(hkey, "EventMessageFile", NULL, &regtype, (LPBYTE) s, &d);
  if (l != ERROR_SUCCESS)
  { d = GetLastError();
    RegCloseKey(hkey);
    SetLastError(d);
    return NULL;
  }

  RegCloseKey(hkey);

// Get the substrings; NB we can only handle 64 of these and there could
// conceivably be more.

  if (m_Record->NumStrings > 64)
    return NULL;

  for (l = 0; l < MAX_SUBSTRINGS; l++)
    substring[l] = "";

  if (m_Record->NumStrings > 0)
  { stringptr = (char*) m_Record + m_Record->StringOffset;

    for (l = 0; l < m_Record->NumStrings && l < MAX_SUBSTRINGS; l++)
    { substring[l] = stringptr;
      stringptr += strlen(stringptr) + 1;
    }
  }

// There can be several message files separated by ;

  msgfileptr = s;

  while (*msgfileptr != '\0')
  { for (l = 0; msgfileptr[l] != ';' && msgfileptr[l] != '\0'; l++);
    if (msgfileptr[l] == ';')
      msgfileptr[l++] = '\0';

    SubstituteEnvironmentVariables(msgfileptr, messagefile, MAX_PATH);

    msgfileptr += l;

// Load the message file and format the message

    h_messagefile = LoadLibraryEx(messagefile, NULL, LOAD_LIBRARY_AS_DATAFILE);
    if (!h_messagefile)
      continue;
  
    l = FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_HMODULE | FORMAT_MESSAGE_ARGUMENT_ARRAY,
                      h_messagefile, 
                      m_Record->EventID, 
                      0, 
                      (LPTSTR) &m_Message, 
                      0, 
                      substring);

    d = GetLastError(); // Save the last error in case the format failed
    FreeLibrary(h_messagefile);
    SetLastError(d);

// If the format succeeded then exit the loop

    if (l != 0)
      break;
  }
  
// All done

  return m_Message;
}


//**********************************************************************
// SubstituteEnvironmentVariables
// ==============================
//**********************************************************************

static DWORD SubstituteEnvironmentVariables(char* InBuf, char* OutBuf, DWORD MaxLen)
{ char env[256], def[256], envval[256];
  char *p_in, *p_out, *p_env, *p_def;
  DWORD maxlen, len, d;

/*** Initialise pointers to start of input and output buffers */

  p_out  = OutBuf;
  p_in   = InBuf;
  maxlen = MaxLen - 1;
  len    = 0;

/*** Loop copying data from inbuf to outbuf */

  while (*p_in != '\0')
  {

/*** If the character is not a % then just copy it */

    if (*p_in != '%')
    { if (len >= maxlen) // Remember to check the length
      { *p_out = '\0';
        return(0);
      }
      else
      { *p_out++ = *p_in++;
        len++;
      }
    }

/*** If the next two characters are %% then copy a single % */

    else if (*(p_in+1) == '%')
    { if (len >= maxlen) // Remember to check the length
      { *p_out = '\0';
        return(0);
      }
      else
      { *p_out++ = *p_in++;
        *p_in++;
        len++;
      }
    }

/*** We have %something so look for an environment variable */

    else
    { p_in++;

/*** Copy everything up to the next % into the env var buffer */

      for (p_env = env; *p_in != '%' && *p_in != '\0'; *p_env++, *p_in++)
        *p_env = *p_in;
      *p_env = '\0';

/*** If a second % was not found just copy the characters to outbuf */

      if (*p_in != '%')
      { if (len >= maxlen) // Remember to check the length
        { *p_out = '\0';
          return(0);
        }
        *p_out++ = '%';
        len++;

        for (p_env = env; *p_env != '\0'; *p_env++, *p_out++)
        { if (len >= maxlen) // Remember to check the length
          { *p_out = '\0';
            return(0);
          }
          else
          { *p_out = *p_env;
          }
        }
      }

/*** OK, we have a substring %xxx% with xxx now in env */

      else
      {

/*** Look for a default character '|' */

        p_def = def;
        *p_def = '\0';

        for (p_env = env; *p_env != '|' && *p_env != '\0'; p_env++);

/*** If there is a '|' then split into env var name and default value */

        if (*p_env == '|')
        { *p_env++ = '\0';
          while (*p_env != '\0')
            *p_def++ = *p_env++;
          *p_def = '\0';
        }

/*** Right, get the env var value */

        d = GetEnvironmentVariable(env, envval, 256);

/*** If the env var was found then copy into outbuf */

        if (d > 0)
        { for (p_env = envval; *p_env!= '\0'; p_env++, p_out++, len++)
          { if (len >= maxlen) // Remember to check the length
            { *p_out = '\0';
              return(0);
            }
            else
            { *p_out = *p_env;
            }
          }
        }

/*** If not found use the default (if any) */

        else if (def[0] != '\0')
        { for (p_def = def; *p_def != '\0'; p_def++, p_out++, len++)
          { if (len >= maxlen) // Remember to check the length
            { *p_out = '\0';
              return(0);
            }
            else
            { *p_out = *p_def;
            }
          }
        }

/*** Finally remember to step past the trailing % */

		p_in++;
      }
    }
  }

/*** Terminate the string */

  *p_out = '\0';

/*** Return the length of the completed string */

  return(len);
}


